# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-28 14:05
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import re
import waldur_core.core.shims
import waldur_azure.validators
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0005_customer_domain'),
        ('core', '0003_enlarge_username'),
        ('taggit', '0002_auto_20150616_2121'),
        ('waldur_azure', '0002_immutable_default_json'),
    ]

    operations = [
        migrations.CreateModel(
            name='Location',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('latitude', models.FloatField(blank=True, null=True)),
                ('longitude', models.FloatField(blank=True, null=True)),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('name', models.CharField(max_length=64, validators=[django.core.validators.RegexValidator(message='The name can contain only letters, numbers, underscore, period and hyphens.', regex=re.compile('[a-zA-Z][a-zA-Z0-9._-]+$'))])),
                ('cidr', models.CharField(max_length=32)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='NetworkInterface',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('name', models.CharField(max_length=80, validators=[django.core.validators.RegexValidator(message='The name can contain only letters, numbers, underscore, period and hyphens.', regex=re.compile('[a-zA-Z][a-zA-Z0-9._-]+$'))])),
                ('config_name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ResourceGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('name', models.CharField(max_length=90, validators=[django.core.validators.RegexValidator(message='The name can include alphanumeric, underscore, parentheses, hyphen, period (except at end), and Unicode characters that match the allowed characters.', regex=re.compile('^[-\\w._()]+$'))])),
                ('location', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.Location')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Size',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('max_data_disk_count', models.PositiveIntegerField()),
                ('memory_in_mb', models.PositiveIntegerField()),
                ('number_of_cores', models.PositiveIntegerField()),
                ('os_disk_size_in_mb', models.PositiveIntegerField()),
                ('resource_disk_size_in_mb', models.PositiveIntegerField()),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SQLDatabase',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('charset', models.CharField(blank=True, max_length=255)),
                ('collation', models.CharField(blank=True, max_length=255)),
                ('resource_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SQLServer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[django.core.validators.RegexValidator(message='The name can only be made up of lowercase letters "a"-"z", the numbers 0-9 and the hyphen. The hyphen may not lead or trail in the name.', regex=re.compile('[a-z0-9][a-z0-9-]+[a-z0-9]$'))], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('username', models.CharField(max_length=50)),
                ('password', models.CharField(max_length=50)),
                ('storage_mb', models.PositiveIntegerField(null=True)),
                ('resource_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SubNet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('name', models.CharField(max_length=80, validators=[django.core.validators.RegexValidator(message='The name can contain only letters, numbers, underscore, period and hyphens.', regex=re.compile('[a-zA-Z][a-zA-Z0-9._-]+$'))])),
                ('cidr', models.CharField(max_length=32)),
                ('network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.Network')),
                ('resource_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.RemoveField(
            model_name='instanceendpoint',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='azureserviceprojectlink',
            name='cloud_service_name',
        ),
        migrations.RemoveField(
            model_name='virtualmachine',
            name='private_ips',
        ),
        migrations.RemoveField(
            model_name='virtualmachine',
            name='public_ips',
        ),
        migrations.RemoveField(
            model_name='virtualmachine',
            name='user_password',
        ),
        migrations.RemoveField(
            model_name='virtualmachine',
            name='user_username',
        ),
        migrations.AddField(
            model_name='image',
            name='offer',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='image',
            name='publisher',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='image',
            name='settings',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='image',
            name='sku',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='image',
            name='version',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='image',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.Image'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='password',
            field=models.CharField(default=None, max_length=72, validators=[django.core.validators.MinLengthValidator(6), django.core.validators.MaxLengthValidator(72), waldur_azure.validators.validate_password]),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='ssh_key',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.SshPublicKey'),
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='username',
            field=models.CharField(default=None, max_length=32, validators=[waldur_azure.validators.VirtualMachineUsernameValidator]),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='image',
            name='backend_id',
            field=models.CharField(db_index=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='virtualmachine',
            name='name',
            field=models.CharField(max_length=15, validators=[django.core.validators.RegexValidator(message='The name can contain only letters, numbers, and hyphens. The name must be shorter than 15 characters and start with a letter and must end with a letter or a number.', regex=re.compile('[a-zA-Z][a-zA-Z0-9-]{0,13}[a-zA-Z0-9]$'))]),
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.DeleteModel(
            name='InstanceEndpoint',
        ),
        migrations.AddField(
            model_name='subnet',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='subnet',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='sqlserver',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='sqlserver',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='sqldatabase',
            name='server',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.SQLServer'),
        ),
        migrations.AddField(
            model_name='sqldatabase',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='sqldatabase',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='resourcegroup',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='resourcegroup',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='networkinterface',
            name='resource_group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup'),
        ),
        migrations.AddField(
            model_name='networkinterface',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='networkinterface',
            name='subnet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.SubNet'),
        ),
        migrations.AddField(
            model_name='networkinterface',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='network',
            name='resource_group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup'),
        ),
        migrations.AddField(
            model_name='network',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='network',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='network_interface',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.NetworkInterface'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='resource_group',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='virtualmachine',
            name='size',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.Size'),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='size',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions
import rest_framework.authentication

import waldur_core.logging.middleware

TOKEN_KEY = settings.WALDUR_CORE.get('TOKEN_KEY', 'x-auth-token')


def can_access_admin_site(user):
    return user.is_active and (user.is_staff or user.is_support)


class AuthenticationBackend:
    """
    Enables only support and staff to access admin site.
    """

    def authenticate(self, request, username, password):
        """
        Always return ``None`` to prevent authentication within this backend.
        """
        return None

    def has_perm(self, user_obj, perm, obj=None):
        return can_access_admin_site(user_obj)

    def has_module_perms(self, user_obj, app_label):
        return can_access_admin_site(user_obj)


class TokenAuthentication(rest_framework.authentication.TokenAuthentication):
    """
    Custom token-based authentication.

    Use TOKEN_KEY from request query parameters if authentication token was not found in header.
    """

    def get_authorization_value(self, request):
        auth = rest_framework.authentication.get_authorization_header(request)
        if not auth:
            auth = request.query_params.get(TOKEN_KEY, '')
        return auth

    def authenticate_credentials(self, key):
        model = self.get_model()
        try:
            token = model.objects.select_related('user').get(key=key)
        except model.DoesNotExist:
            raise exceptions.AuthenticationFailed(_('Invalid token.'))

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed(_('User inactive or deleted.'))

        if token.user.token_lifetime:
            lifetime = timezone.timedelta(seconds=token.user.token_lifetime)

            if token.created < timezone.now() - lifetime:
                raise exceptions.AuthenticationFailed(_('Token has expired.'))

        return token.user, token

    def authenticate(self, request):
        auth = self.get_authorization_value(request).split()

        if not auth or auth[0].lower() != b'token':
            return None

        if len(auth) == 1:
            msg = _('Invalid token. No credentials provided.')
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _('Invalid token. Token string should not contain spaces.')
            raise exceptions.AuthenticationFailed(msg)

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = _('Invalid token header. Token string should not contain invalid characters.')
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(token)


def user_capturing_auth(auth):
    class CapturingAuthentication(auth):
        def authenticate(self, request):
            result = super(CapturingAuthentication, self).authenticate(request)
            if result is not None:
                user, _ = result
                waldur_core.logging.middleware.set_current_user(user)
                token = user.auth_token
                if token:
                    token.created = timezone.now()
                    token.save()
            return result

    return CapturingAuthentication


class CsrfExemptSessionAuthentication(rest_framework.authentication.SessionAuthentication):

    def enforce_csrf(self, request):
        return  # Skip CSRF check


SessionAuthentication = user_capturing_auth(CsrfExemptSessionAuthentication)
TokenAuthentication = user_capturing_auth(TokenAuthentication)



# -*- coding: utf-8 -*-
from decimal import Decimal
from django.conf import settings
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators
import waldur_core.logging.loggers
import waldur_core.media.models
import waldur_core.media.validators
import waldur_core.structure.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AggregateResourceCount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('count', models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('key', models.CharField(max_length=255, primary_key=True, serialize=False, validators=[django.core.validators.RegexValidator('^[a-zA-Z0-9_]+$')])),
                ('title', models.CharField(max_length=255)),
                ('type', models.CharField(choices=[('boolean', 'boolean'), ('string', 'string'), ('text', 'text'), ('integer', 'integer'), ('choice', 'choice'), ('list', 'list')], max_length=255)),
                ('required', models.BooleanField(default=False, help_text='A value must be provided for the attribute.')),
                ('default', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AttributeOption',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=255, validators=[django.core.validators.RegexValidator('^[a-zA-Z0-9_]+$')])),
                ('title', models.CharField(max_length=255)),
                ('attribute', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='options', to='marketplace.Attribute')),
            ],
        ),
        migrations.CreateModel(
            name='CartItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
                ('limits', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('type', models.PositiveSmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Terminate')], default=1)),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
            ],
            options={
                'ordering': ('created',),
            },
        ),
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('title', models.CharField(max_length=255)),
                ('icon', models.FileField(blank=True, null=True, upload_to='marketplace_category_icons', validators=[waldur_core.media.validators.FileTypeValidator(allowed_types=['image/png', 'image/gif', 'image/jpeg', 'image/svg', 'image/svg+xml', 'image/x-icon'])])),
                ('description', models.TextField(blank=True)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'ordering': ('title',),
                'verbose_name': 'Category',
                'verbose_name_plural': 'Categories',
            },
        ),
        migrations.CreateModel(
            name='CategoryColumn',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.PositiveSmallIntegerField(help_text='Index allows to reorder columns.')),
                ('title', models.CharField(help_text='Title is rendered as column header.', max_length=255)),
                ('attribute', models.CharField(blank=True, help_text='Resource attribute is rendered as table cell.', max_length=255)),
                ('widget', models.CharField(blank=True, help_text='Widget field allows to customise table cell rendering.', max_length=255)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='columns', to='marketplace.Category')),
            ],
            options={
                'ordering': ('category', 'index'),
            },
        ),
        migrations.CreateModel(
            name='CategoryComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(help_text='Display name for the measured unit, for example, Floating IP.', max_length=150)),
                ('type', models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50, validators=[django.core.validators.RegexValidator('^[a-zA-Z0-9_]+$')])),
                ('measured_unit', models.CharField(blank=True, help_text='Unit of measurement, for example, GB.', max_length=30)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Category')),
            ],
        ),
        migrations.CreateModel(
            name='CategoryComponentUsage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('date', models.DateField()),
                ('reported_usage', models.PositiveIntegerField(null=True)),
                ('fixed_usage', models.PositiveIntegerField(null=True)),
                ('component', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.CategoryComponent')),
                ('content_type', models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ComponentQuota',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('limit', models.PositiveIntegerField(default=-1)),
                ('usage', models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='ComponentUsage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('usage', models.PositiveIntegerField(default=0)),
                ('date', models.DateTimeField()),
                ('billing_period', models.DateField()),
            ],
        ),
        migrations.CreateModel(
            name='Offering',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('thumbnail', models.FileField(blank=True, null=True, upload_to='marketplace_service_offering_thumbnails', validators=[waldur_core.media.validators.FileTypeValidator(allowed_types=['image/png', 'image/gif', 'image/jpeg', 'image/svg', 'image/svg+xml', 'image/x-icon'])])),
                ('full_description', models.TextField(blank=True)),
                ('vendor_details', models.TextField(blank=True)),
                ('rating', models.IntegerField(help_text='Rating is value from 1 to 5.', null=True, validators=[django.core.validators.MaxValueValidator(5), django.core.validators.MinValueValidator(1)])),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Fields describing Category.')),
                ('options', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Fields describing Offering request form.')),
                ('geolocations', waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of latitudes and longitudes. For example: [{"latitude": 123, "longitude": 345}, {"latitude": 456, "longitude": 678}]')),
                ('native_name', models.CharField(blank=True, default='', max_length=160)),
                ('native_description', models.CharField(blank=True, default='', max_length=500)),
                ('terms_of_service', models.TextField(blank=True)),
                ('type', models.CharField(max_length=100)),
                ('state', django_fsm.FSMIntegerField(choices=[(1, 'Draft'), (2, 'Active'), (3, 'Paused'), (4, 'Archived')], default=1)),
                ('paused_reason', models.TextField(blank=True)),
                ('shared', models.BooleanField(default=True, help_text='Accessible to all customers.')),
                ('billable', models.BooleanField(default=True, help_text='Purchase and usage is invoiced.')),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('allowed_customers', models.ManyToManyField(blank=True, to='structure.Customer')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offerings', to='marketplace.Category')),
                ('content_type', models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('customer', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Customer')),
            ],
            options={
                'verbose_name': 'Offering',
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.CreateModel(
            name='OfferingComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('name', models.CharField(help_text='Display name for the measured unit, for example, Floating IP.', max_length=150)),
                ('type', models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50, validators=[django.core.validators.RegexValidator('^[a-zA-Z0-9_]+$')])),
                ('measured_unit', models.CharField(blank=True, help_text='Unit of measurement, for example, GB.', max_length=30)),
                ('billing_type', models.CharField(choices=[('fixed', 'Fixed-price'), ('usage', 'Usage-based'), ('one', 'One-time'), ('few', 'One-time on plan switch')], default='fixed', max_length=5)),
                ('limit_period', models.CharField(blank=True, choices=[('month', 'Maximum monthly - every month service provider can report up to the amount requested by user.'), ('total', 'Maximum total - SP can report up to the requested amount over the whole active state of resource.')], max_length=5, null=True)),
                ('limit_amount', models.IntegerField(blank=True, null=True)),
                ('max_value', models.IntegerField(blank=True, null=True)),
                ('min_value', models.IntegerField(blank=True, null=True)),
                ('disable_quotas', models.BooleanField(default=False, help_text='Do not allow user to specify quotas when offering is provisioned.')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Offering')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.CategoryComponent')),
            ],
        ),
        migrations.CreateModel(
            name='OfferingFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('file', models.FileField(upload_to='offering_files')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='files', to='marketplace.Offering')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('approved_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(1, 'requested for approval'), (2, 'executing'), (3, 'done'), (4, 'terminated'), (5, 'erred'), (6, 'rejected')], default=1)),
                ('total_cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
                ('_file', models.TextField(blank=True, editable=False)),
                ('approved_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='orders', to=settings.AUTH_USER_MODEL)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
            ],
            options={
                'ordering': ('created',),
                'verbose_name': 'Order',
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.CreateModel(
            name='OrderItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
                ('limits', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('type', models.PositiveSmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Terminate')], default=1)),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('state', django_fsm.FSMIntegerField(choices=[(1, 'pending'), (2, 'executing'), (3, 'done'), (4, 'erred'), (5, 'terminated'), (6, 'terminating')], default=1)),
                ('activated', models.DateTimeField(blank=True, null=True, verbose_name='activation date')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.Offering')),
            ],
            options={
                'ordering': ('created',),
                'verbose_name': 'Order item',
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[('month', 'Per month'), ('half_month', 'Per half month'), ('day', 'Per day'), ('hour', 'Per hour'), ('quantity', 'Quantity')], default='day', max_length=30)),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('archived', models.BooleanField(default=False, help_text='Forbids creation of new resources.')),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('max_amount', models.PositiveSmallIntegerField(blank=True, help_text='Maximum number of plans that could be active. Plan is disabled when maximum amount is reached.', null=True, validators=[django.core.validators.MinValueValidator(1)])),
                ('content_type', models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plans', to='marketplace.Offering')),
            ],
            options={
                'ordering': ('name',),
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.CreateModel(
            name='PlanComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.PositiveIntegerField(default=0)),
                ('price', models.DecimalField(decimal_places=7, default=0, max_digits=15, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per unit per billing period.')),
                ('component', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.OfferingComponent')),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Plan')),
            ],
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('cost', models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True)),
                ('limits', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('state', django_fsm.FSMIntegerField(choices=[(1, 'Creating'), (2, 'OK'), (3, 'Erred'), (4, 'Updating'), (5, 'Terminating'), (6, 'Terminated')], default=1)),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('backend_metadata', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('current_usages', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
                ('content_type', models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='marketplace.Offering')),
                ('plan', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ResourcePlanPeriod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('start', models.DateTimeField(blank=True, null=True, verbose_name='start')),
                ('end', models.DateTimeField(blank=True, null=True, verbose_name='end')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan')),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Resource')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Screenshot',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('image', models.ImageField(upload_to=waldur_core.media.models.get_upload_path)),
                ('thumbnail', models.ImageField(editable=False, null=True, upload_to=waldur_core.media.models.get_upload_path)),
                ('offering', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='screenshots', to='marketplace.Offering')),
            ],
            options={
                'verbose_name': 'Screenshot',
            },
        ),
        migrations.CreateModel(
            name='Section',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('key', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=255)),
                ('is_standalone', models.BooleanField(default=False, help_text='Whether section is rendered as a separate tab.')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sections', to='marketplace.Category')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ServiceProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('enable_notifications', models.BooleanField(default=True)),
                ('api_secret_code', models.CharField(blank=True, max_length=255, null=True)),
                ('lead_email', models.EmailField(blank=True, help_text='Email for notification about new request based order items. If this field is set, notifications will be sent.', max_length=254, null=True)),
                ('lead_subject', models.CharField(blank=True, help_text='Notification subject template. Django template variables can be used.', max_length=255)),
                ('lead_body', models.TextField(blank=True, help_text='Notification body template. Django template variables can be used.')),
                ('customer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer')),
            ],
            options={
                'verbose_name': 'Service provider',
            },
        ),
        migrations.AddField(
            model_name='orderitem',
            name='old_plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Plan'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='order',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='marketplace.Order'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='resource',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Resource'),
        ),
        migrations.AddField(
            model_name='componentusage',
            name='component',
            field=models.ForeignKey(limit_choices_to={'billing_type': 'usage'}, on_delete=django.db.models.deletion.CASCADE, to='marketplace.OfferingComponent'),
        ),
        migrations.AddField(
            model_name='componentusage',
            name='plan_period',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.ResourcePlanPeriod'),
        ),
        migrations.AddField(
            model_name='componentusage',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='usages', to='marketplace.Resource'),
        ),
        migrations.AddField(
            model_name='componentquota',
            name='component',
            field=models.ForeignKey(limit_choices_to={'billing_type': 'usage'}, on_delete=django.db.models.deletion.CASCADE, to='marketplace.OfferingComponent'),
        ),
        migrations.AddField(
            model_name='componentquota',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='quotas', to='marketplace.Resource'),
        ),
        migrations.AddField(
            model_name='cartitem',
            name='offering',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Offering'),
        ),
        migrations.AddField(
            model_name='cartitem',
            name='plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan'),
        ),
        migrations.AddField(
            model_name='cartitem',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='cartitem',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='attribute',
            name='section',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='marketplace.Section'),
        ),
        migrations.AddField(
            model_name='aggregateresourcecount',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Category'),
        ),
        migrations.AddField(
            model_name='aggregateresourcecount',
            name='content_type',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
        migrations.AlterUniqueTogether(
            name='plancomponent',
            unique_together=set([('plan', 'component')]),
        ),
        migrations.AlterUniqueTogether(
            name='offeringcomponent',
            unique_together=set([('type', 'offering')]),
        ),
        migrations.AlterUniqueTogether(
            name='componentusage',
            unique_together=set([('resource', 'component', 'plan_period', 'billing_period')]),
        ),
        migrations.AlterUniqueTogether(
            name='componentquota',
            unique_together=set([('resource', 'component')]),
        ),
        migrations.AlterUniqueTogether(
            name='categorycomponent',
            unique_together=set([('type', 'category')]),
        ),
        migrations.AlterUniqueTogether(
            name='attributeoption',
            unique_together=set([('attribute', 'key')]),
        ),
        migrations.AlterUniqueTogether(
            name='aggregateresourcecount',
            unique_together=set([('category', 'content_type', 'object_id')]),
        ),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger
from waldur_core.structure.models import Project
from waldur_core.structure.permissions import _get_project

from . import models


def get_issue_scopes(issue):
    result = set()
    if issue.resource:
        project = _get_project(issue.resource)
        result.add(issue.resource)
        result.add(project)
        result.add(project.customer)
    if issue.project:
        result.add(issue.project)
        result.add(issue.customer)
    if issue.customer:
        result.add(issue.customer)
    return result


class IssueEventLogger(EventLogger):
    issue = models.Issue

    class Meta:
        event_types = ('issue_deletion_succeeded',
                       'issue_update_succeeded',
                       'issue_creation_succeeded')
        event_groups = {
            'support': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        issue = event_context['issue']
        return get_issue_scopes(issue)


class AttachmentEventLogger(EventLogger):
    attachment = models.Attachment

    class Meta:
        event_types = (
            'attachment_created',
            'attachment_updated',
            'attachment_deleted',
        )
        event_groups = {
            'support': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        attachment = event_context['attachment']
        return get_issue_scopes(attachment.issue)


class OfferingEventLogger(EventLogger):
    offering = models.Offering

    class Meta:
        event_types = (
            'offering_created',
            'offering_deleted',
            'offering_state_changed',
        )
        event_groups = {
            'support': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        offering = event_context['offering']
        project = Project.all_objects.get(id=offering.project_id)  # handle case when project is already deleted
        return {offering, project, project.customer}


event_logger.register('waldur_issue', IssueEventLogger)
event_logger.register('waldur_attachment', AttachmentEventLogger)
event_logger.register('waldur_offering', OfferingEventLogger)



# Generated by Django 2.2.7 on 2020-01-10 12:10

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_extend_organization'),
    ]

    operations = [
        migrations.CreateModel(
            name='ChangeEmailRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('email', models.EmailField(max_length=254)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'change email request',
                'verbose_name_plural': 'change email requests',
            },
        ),
    ]



# Generated by Django 2.2.7 on 2020-01-20 10:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0009_project_is_removed'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='latitude',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='longitude',
            field=models.FloatField(blank=True, null=True),
        ),
    ]



from django.apps import AppConfig


class MarketplaceScriptConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_script'
    verbose_name = 'Marketplace Script'

    def ready(self):
        from waldur_mastermind.marketplace.plugins import manager

        from . import PLUGIN_NAME, processors

        manager.register(
            offering_type=PLUGIN_NAME,
            create_resource_processor=processors.CreateProcessor,
            update_resource_processor=processors.UpdateProcessor,
            delete_resource_processor=processors.DeleteProcessor,
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-29 16:37
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_azure', '0005_ordering'),
    ]

    operations = [
        migrations.AlterField(
            model_name='virtualmachine',
            name='user_data',
            field=models.TextField(blank=True, help_text='Additional data that will be added to instance on provisioning', max_length=87380),
        ),
    ]



from django.conf import settings
from django.db import models
from rest_framework import serializers

from waldur_core.media.utils import encode_protected_url
from waldur_core.structure.metadata import merge_dictionaries


class ProtectedFileMixin:
    def to_representation(self, value):
        if not value:
            return None

        if settings.DEBUG:
            return super(ProtectedFileMixin, self).to_representation(value)

        return encode_protected_url(value.instance,
                                    field=self.source_attrs[-1],
                                    request=self.context['request'])


class ProtectedFileField(ProtectedFileMixin, serializers.FileField):
    pass


class ProtectedImageField(ProtectedFileMixin, serializers.ImageField):
    pass


class ProtectedMediaSerializerMixin(serializers.ModelSerializer):
    serializer_field_mapping = merge_dictionaries(
        serializers.ModelSerializer.serializer_field_mapping,
        {
            models.FileField: ProtectedFileField,
            models.ImageField: ProtectedImageField,
        },
    )



from . import views

from django.conf.urls import url

urlpatterns = [
    url(r'^api/marketplace-checklists/$',
        views.ChecklistView.as_view({'get': 'list'})),
    url(r'^api/marketplace-checklists/(?P<checklist_uuid>[a-f0-9]+)/questions/$',
        views.QuestionsView.as_view({'get': 'list'})),
    url(r'^api/marketplace-checklists/(?P<checklist_uuid>[a-f0-9]+)/stats/$',
        views.StatsView.as_view()),
    url(r'^api/projects/(?P<project_uuid>[a-f0-9]+)/marketplace-checklists/$',
        views.ProjectStatsView.as_view()),
    url(r'^api/marketplace-checklists/(?P<checklist_uuid>[a-f0-9]+)/answers/(?P<project_uuid>[a-z0-9]+)/$',
        views.AnswersListView.as_view({'get': 'list'})),
    url(r'^api/marketplace-checklists/(?P<checklist_uuid>[a-f0-9]+)/answers/(?P<project_uuid>[a-z0-9]+)/submit/$',
        views.AnswersSubmitView.as_view({'post': 'create'})),
]



import collections
import logging

from django.db import models
from django.utils.lru_cache import lru_cache
from django.utils.topological_sort import stable_topological_sort
from django.utils.translation import ugettext_lazy as _
from rest_framework.exceptions import ValidationError

from . import SupportedServices

logger = logging.getLogger(__name__)
FieldInfo = collections.namedtuple('FieldInfo', 'fields fields_required extra_fields_required extra_fields_default')


@lru_cache(maxsize=1)
def get_sorted_dependencies(service_model):
    """
    Returns list of application models in topological order.
    It is used in order to correctly delete dependent resources.
    """
    app_models = list(service_model._meta.app_config.get_models())
    dependencies = {model: set() for model in app_models}
    relations = (
        relation
        for model in app_models
        for relation in model._meta.related_objects
        if relation.on_delete in (models.PROTECT, models.CASCADE)
    )
    for rel in relations:
        dependencies[rel.model].add(rel.related_model)
    return stable_topological_sort(app_models, dependencies)


def sort_dependencies(service_model, resources):
    ordering = get_sorted_dependencies(service_model)
    resources.sort(key=lambda resource: ordering.index(resource._meta.model))
    return resources


@lru_cache(maxsize=1)
def get_all_services_field_info():
    services_fields = dict()
    services_fields_default_value = dict()
    services_fields_required = dict()
    services_extra_fields_required = dict()
    service_models = SupportedServices.get_service_models()

    for service_name in service_models:
        service_model = service_models[service_name]['service']
        service_serializer = SupportedServices.get_service_serializer(service_model)

        fields = service_serializer.SERVICE_ACCOUNT_FIELDS.keys() \
            if service_serializer.SERVICE_ACCOUNT_FIELDS is not NotImplemented else []

        extra_field_options = getattr(service_serializer.Meta, 'extra_field_options', {})
        fields_default_value = {k: v.get('default_value')
                                for k, v in extra_field_options.items()
                                if v.get('default_value')}

        fields_extra = service_serializer.SERVICE_ACCOUNT_EXTRA_FIELDS.keys() \
            if service_serializer.SERVICE_ACCOUNT_EXTRA_FIELDS is not NotImplemented else []

        fields_required = service_serializer.Meta.required_fields \
            if hasattr(service_serializer.Meta, 'required_fields') else []

        services_fields[service_name] = list(fields)
        services_fields_required[service_name] = list(set(fields) & set(fields_required))
        services_extra_fields_required[service_name] = list(set(fields_extra) & set(fields_required))
        services_fields_default_value[service_name] = fields_default_value

    return FieldInfo(fields=services_fields,
                     fields_required=services_fields_required,
                     extra_fields_required=services_extra_fields_required,
                     extra_fields_default=services_fields_default_value)


def update_pulled_fields(instance, imported_instance, fields):
    """
    Update instance fields based on imported from backend data.
    Save changes to DB only one or more fields were changed.
    """
    modified = False
    for field in fields:
        pulled_value = getattr(imported_instance, field)
        current_value = getattr(instance, field)
        if current_value != pulled_value:
            setattr(instance, field, pulled_value)
            logger.info("%s's with PK %s %s field updated from value '%s' to value '%s'",
                        instance.__class__.__name__, instance.pk, field, current_value, pulled_value)
            modified = True
    error_message = getattr(imported_instance, 'error_message', '') or getattr(instance, 'error_message', '')
    if error_message and instance.error_message != error_message:
        instance.error_message = imported_instance.error_message
        modified = True
    if modified:
        instance.save()


def handle_resource_not_found(resource):
    """
    Set resource state to ERRED and append/create "not found" error message.
    """
    resource.set_erred()
    resource.runtime_state = ''
    message = 'Does not exist at backend.'
    if message not in resource.error_message:
        if not resource.error_message:
            resource.error_message = message
        else:
            resource.error_message += ' (%s)' % message
    resource.save()
    logger.warning('%s %s (PK: %s) does not exist at backend.' % (
        resource.__class__.__name__, resource, resource.pk))


def handle_resource_update_success(resource):
    """
    Recover resource if its state is ERRED and clear error message.
    """
    update_fields = []
    if resource.state == resource.States.ERRED:
        resource.recover()
        update_fields.append('state')

    if resource.state in (resource.States.UPDATING, resource.States.CREATING):
        resource.set_ok()
        update_fields.append('state')

    if resource.error_message:
        resource.error_message = ''
        update_fields.append('error_message')

    if update_fields:
        resource.save(update_fields=update_fields)
    logger.info('%s %s (PK: %s) was successfully updated.' % (
        resource.__class__.__name__, resource, resource.pk))


def check_customer_blocked(obj):
    from waldur_core.structure import permissions

    customer = permissions._get_customer(obj)
    if customer and customer.blocked:
        raise ValidationError(_('Blocked organization is not available.'))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-25 13:05
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.shims
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.structure.models


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('waldur_vmware', '0002_virtualmachine'),
    ]

    operations = [
        migrations.CreateModel(
            name='Disk',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('size', models.PositiveIntegerField(help_text='Size in MiB')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='waldur_vmware.VMwareServiceProjectLink')),
                ('tags', waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('vm', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='disks', to='waldur_vmware.VirtualMachine')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
    ]



import datetime

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models as django_models
from django.db.models import Q
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_core.core.managers import GenericKeyMixin
from waldur_core.structure.managers import filter_queryset_for_user
from waldur_core.structure.models import Service


# TODO: This mixin duplicates quota filter manager - they need to be moved to core (NC-686)
class UserFilterMixin:

    def filtered_for_user(self, user, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        if user.is_staff or user.is_support:
            return queryset

        query = Q()
        for model in self.get_available_models():
            user_object_ids = filter_queryset_for_user(model.objects.all(), user).values_list('id', flat=True)
            content_type_id = ContentType.objects.get_for_model(model).id
            query |= Q(object_id__in=list(user_object_ids), content_type_id=content_type_id)

        return queryset.filter(query)

    def get_available_models(self):
        """ Return list of models that are acceptable """
        raise NotImplementedError()


class PriceEstimateManager(GenericKeyMixin, UserFilterMixin, django_models.Manager):

    def get_available_models(self):
        """ Return list of models that are acceptable """
        return self.model.get_estimated_models()

    def get_current(self, scope):
        now = timezone.now()
        return self.get(scope=scope, year=now.year, month=now.month)

    def get_or_create_current(self, scope):
        now = timezone.now()
        return self.get_or_create(scope=scope, month=now.month, year=now.year)

    def filter_current(self):
        now = timezone.now()
        return self.filter(year=now.year, month=now.month)


class ConsumptionDetailsQuerySet(django_models.QuerySet):

    def create(self, price_estimate):
        """ Take configuration from previous month, it it exists.
            Set last_update_time equals to the beginning of the month.
        """
        kwargs = {}
        try:
            previous_price_estimate = price_estimate.get_previous()
        except ObjectDoesNotExist:
            pass
        else:
            configuration = previous_price_estimate.consumption_details.configuration
            kwargs['configuration'] = configuration
        month_start = core_utils.month_start(datetime.date(price_estimate.year, price_estimate.month, 1))
        kwargs['last_update_time'] = month_start
        return super(ConsumptionDetailsQuerySet, self).create(price_estimate=price_estimate, **kwargs)


ConsumptionDetailsManager = django_models.Manager.from_queryset(ConsumptionDetailsQuerySet)


class PriceListItemManager(GenericKeyMixin, UserFilterMixin, django_models.Manager):

    def get_available_models(self):
        """ Return list of models that are acceptable """
        return Service.get_all_models()



from django.apps import AppConfig
from django.db.models import signals


class MarketplaceVMwareConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_vmware'
    verbose_name = 'Marketplace VMware'

    def ready(self):
        from waldur_mastermind.marketplace.plugins import Component
        from waldur_mastermind.marketplace.plugins import manager
        from waldur_mastermind.marketplace import handlers as marketplace_handlers
        from waldur_mastermind.invoices import registrators
        from waldur_mastermind.marketplace import models as marketplace_models
        from waldur_vmware import models as vmware_models
        from waldur_vmware import signals as vmware_signals
        from waldur_vmware.apps import VMwareConfig

        from . import handlers, registrators as vmware_registrators, processors, VIRTUAL_MACHINE_TYPE

        resource_models = (
            vmware_models.VirtualMachine,
        )

        marketplace_handlers.connect_resource_handlers(*resource_models)
        marketplace_handlers.connect_resource_metadata_handlers(*resource_models)

        USAGE = marketplace_models.OfferingComponent.BillingTypes.USAGE
        manager.register(offering_type=VIRTUAL_MACHINE_TYPE,
                         create_resource_processor=processors.VirtualMachineCreateProcessor,
                         service_type=VMwareConfig.service_name,
                         components=(
                             Component(type='cpu', name='CPU', measured_unit='vCPU', billing_type=USAGE),
                             # Price is stored per GiB but size is stored per MiB
                             # therefore we need to divide size by factor when price estimate is calculated.
                             Component(type='ram', name='RAM', measured_unit='GB', billing_type=USAGE, factor=1024),
                             Component(type='disk', name='Disk', measured_unit='GB', billing_type=USAGE, factor=1024),
                         ))

        registrators.RegistrationManager.add_registrator(
            vmware_models.VirtualMachine,
            vmware_registrators.VirtualMachineRegistrator
        )

        vmware_signals.vm_created.connect(
            handlers.add_new_vm_to_invoice,
            dispatch_uid='marketplace_vmware.handlers.add_new_vm_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.terminate_invoice_when_vm_deleted,
            sender=vmware_models.VirtualMachine,
            dispatch_uid='marketplace_vmware.handlers.terminate_invoice_when_vm_deleted',
        )

        vmware_signals.vm_updated.connect(
            handlers.create_invoice_item_when_vm_is_updated,
            dispatch_uid='marketplace_vmware.handlers.create_invoice_item_when_vm_is_updated',
        )

        vmware_signals.vm_updated.connect(
            handlers.update_marketplace_resource_limits_when_vm_is_updated,
            dispatch_uid='marketplace_vmware.handlers.'
                         'update_marketplace_resource_limits_when_vm_is_updated',
        )



# Generated by Django 2.2.7 on 2019-11-26 11:36

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0005_node_name_is_unique'),
    ]

    operations = [
        migrations.AddField(
            model_name='node',
            name='initial_data',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Initial data for instance creating.'),
        ),
    ]



import contextlib
import json
import logging

from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand
from django.db import transaction
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIClient

from waldur_core.structure import models as structure_models
from waldur_core.structure.tests import factories as structure_factories
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace.tests import factories
from waldur_mastermind.marketplace.utils import process_order_item
from waldur_mastermind.support import models as support_models

logger = logging.getLogger(__name__)


class OrderTestException(Exception):
    pass


class OrderTest:
    def __init__(self, options, stdout_success):
        self.client = APIClient()
        self.stdout = stdout_success

        self.action = options.get('action')
        organization_name = options.get('organization')
        offering_uuid = options.get('offering')
        attributes_path = options.get('attributes')
        self.attributes = {}

        if attributes_path:
            with open(attributes_path) as attributes_file:
                attributes = attributes_file.read()
                self.attributes = json.loads(attributes)

        try:
            self.offering = marketplace_models.Offering.objects.get(uuid=offering_uuid)
        except marketplace_models.Offering.DoesNotExist:
            raise OrderTestException('Offering does not exist.')

        try:
            self.customer = structure_models.Customer.objects.get(name=organization_name)
            self.project = structure_models.Project.objects.get(customer=self.customer)
            self.owner = self.customer.get_owners().first()
            self.admin = self.project.get_users(role=structure_models.ProjectRole.ADMINISTRATOR).first()
        except structure_models.Customer.DoesNotExist:
            self.customer = structure_factories.CustomerFactory(name=organization_name)
            self.owner = structure_factories.UserFactory(username='test owner')
            self.customer.add_user(self.owner, structure_models.CustomerRole.OWNER)
            self.project = structure_factories.ProjectFactory(customer=self.customer)
            self.admin = structure_factories.UserFactory(username='test admin')
            self.project.add_user(self.admin, structure_models.ProjectRole.ADMINISTRATOR)
            self.stdout('Organization is created.')

        for i in range(1, 2 - self.offering.plans.count()):
            factories.PlanFactory(offering=self.offering)

        self.plan_1 = self.offering.plans.first()
        self.plan_2 = self.offering.plans.last()

    @property
    def order_item(self):
        order_item_queryset = marketplace_models.OrderItem.objects.filter(offering=self.offering,
                                                                          order__project=self.project)
        if order_item_queryset.exists():
            return order_item_queryset.last()

    @property
    def order(self):
        if self.order_item:
            return self.order_item.order

    @contextlib.contextmanager
    def mute_stdout(self):
        stdout = self.stdout
        self.stdout = lambda x: x
        try:
            yield {}
        finally:
            self.stdout = stdout

    def get_request_type(self):
        choices = marketplace_models.RequestTypeMixin.Types.CHOICES
        return filter(lambda x: x[0] == self.order_item.type, choices)[0][1].lower()

    def get_issue(self):
        if self.get_request_type() == 'create':
            return self.order_item.resource.scope.issue
        else:
            ct = ContentType.objects.get_for_model(marketplace_models.OrderItem)
            return support_models.Issue.objects.get(resource_content_type=ct,
                                                    resource_object_id=self.order_item.pk)

    def get_response(self, user, url_name, data=None, action=None, uuid=None, status_code=status.HTTP_200_OK):
        if uuid:
            url = 'http://localhost%s' % reverse(url_name, kwargs={'uuid': uuid})
        else:
            url = 'http://localhost%s' % reverse(url_name)

        if action:
            url += action + '/'

        self.client.force_authenticate(user)
        response = self.client.post(url, data)

        if response.status_code != status_code:
            raise OrderTestException('Request %s failed. %s' % (action, response.rendered_content))

        return response

    def approve_order(self):
        self.get_response(self.owner, 'marketplace-order-detail', action='approve', uuid=self.order.uuid)
        process_order_item(self.order_item, self.owner)

        self.stdout('A %s order has been approved.' % self.get_request_type())
        self.stdout('Order UUID: %s' % self.order.uuid)
        self.stdout('Request UUID: %s' % self.order_item.resource.scope.uuid)
        issue = self.get_issue()
        self.stdout('Issue UUID: %s, PK: %s' % (issue.uuid.hex, issue.pk))

    def validate_order_done(self):
        if self.order.state != marketplace_models.Order.States.DONE:
            raise OrderTestException('An order is not done.')

    def validate_request_state(self, issue_resolved, request_state, request_type):
        if request_type == 'create' and issue_resolved and request_state == support_models.Offering.States.OK:
            return

        if request_type == 'create' and not issue_resolved and \
                request_state == support_models.Offering.States.TERMINATED:
            return

        if request_type == 'terminate' and issue_resolved and request_state is None:
            return

        if request_type == 'terminate' and not issue_resolved and request_state == support_models.Offering.States.OK:
            return

        if request_type == 'update' and request_state == support_models.Offering.States.OK:
            return

        raise OrderTestException('Request state is wrong.')

    def issue_info(self):
        if self.order.state == marketplace_models.Order.States.EXECUTING:
            self.stdout('STEP 2: resolve or cancel an issue.')
            self.stdout('A %s order UUID: %s' % (self.get_request_type(), self.order.uuid))
            self.stdout('Request UUID: %s' % self.order_item.resource.scope.uuid)
            issue = self.get_issue()
            self.stdout('Issue UUID: %s, PK: %s' % (issue.uuid.hex, issue.pk))
            self.stdout('Please, resolve or cancel an issue.')

        if self.order.state == marketplace_models.Order.States.DONE:
            issue = self.get_issue()

            if issue.resolved is None:
                raise OrderTestException('An order is done, but the issue is not resolved or canceled.')
            elif issue.resolved:
                self.stdout('FINISH: A %s order has been resolved.' % self.get_request_type())
            else:
                self.stdout('FINISH: A %s order has been canceled.' % self.get_request_type())

            if self.order_item.resource.scope:
                self.stdout('Request state is: %s.' % self.order_item.resource.scope.state)
            else:
                self.stdout('Request with ID %s has been deleted.' % self.order_item.resource.object_id)

            self.stdout('Resource plan is: %s' % self.order_item.resource.plan.name)

            request_state = self.order_item.resource.scope and self.order_item.resource.scope.state
            self.validate_request_state(issue.resolved, request_state, self.get_request_type())

        if self.order.state == marketplace_models.Order.States.REJECTED:
            self.stdout('A %s order has been rejected.' % self.get_request_type())

        if self.order.state == marketplace_models.Order.States.ERRED:
            self.stdout('A %s order has failed.' % self.get_request_type())

    def create_request(self):
        self.stdout('STEP 1: Make an order.')
        data = {
            'project': 'http://localhost' + reverse('project-detail', kwargs={'uuid': self.project.uuid}),
            'items': [
                {
                    'offering': 'http://localhost' + reverse('marketplace-offering-detail',
                                                             kwargs={'uuid': self.offering.uuid}),
                    'attributes': self.attributes,
                    'limits': {},
                    'plan': 'http://localhost' + reverse('marketplace-plan-detail',
                                                         kwargs={'uuid': self.plan_1.uuid}),
                },
            ]
        }

        self.get_response(self.admin, 'marketplace-order-list', data=data, status_code=status.HTTP_201_CREATED)
        self.approve_order()

    def choice_step(func):
        def wrapped(self):
            if self.order_item:
                self.issue_info()
            else:
                func(self)

        return wrapped

    @choice_step
    def create(self):
        self.create_request()

    @choice_step
    def terminate(self):
        self.stdout('STEP 1: Make an order to terminate a resource.')
        with self.mute_stdout():
            self.create_request()
            self.order_item.resource.scope.issue.set_resolved()

        self.validate_order_done()
        self.get_response(self.admin, 'marketplace-resource-detail', action='terminate',
                          uuid=self.order_item.resource.uuid)
        self.approve_order()

    @choice_step
    def update(self):
        self.stdout('STEP 1: Make an order to switch plan of a resource.')
        with self.mute_stdout():
            self.create_request()
            self.order_item.resource.scope.issue.set_resolved()

        self.validate_order_done()
        self.stdout('Resource plan is: %s' % self.order_item.resource.plan.name)
        data = {
            'plan': 'http://localhost' + reverse('marketplace-plan-detail',
                                                 kwargs={'uuid': self.plan_2.uuid}),
        }
        self.get_response(self.admin, 'marketplace-resource-detail', data=data,
                          action='switch_plan', uuid=self.order_item.resource.uuid)

        self.approve_order()

    def delete(self):
        order_item_queryset = marketplace_models.OrderItem.objects.filter(offering=self.offering,
                                                                          order__project=self.project)
        if not order_item_queryset.exists():
            self.stdout('The order has already been deleted.')
            return

        for order_item in order_item_queryset.all():
            order_item.order.delete()

        self.stdout('Order has been deleted.')

    def run(self):
        getattr(self, self.action, lambda: None)()


class Command(BaseCommand):
    help = """Validate capability of a Marketplace plugin."""

    def add_arguments(self, parser):
        parser.add_argument('--offering', type=str, help='Offering UUID', required=True)
        parser.add_argument('--organization', type=str, help='Organization name', default='Test organization')
        parser.add_argument('--action', type=str, help='Action', default='create',
                            choices=['create', 'delete', 'terminate', 'update'])
        parser.add_argument('--attributes', type=str, help='Path to JSON file with order item attributes.', default='')

    def handle(self, *args, **options):
        try:
            with transaction.atomic():
                OrderTest(options, lambda m: self.stdout.write(self.style.SUCCESS(m))).run()
        except OrderTestException as e:
            self.stdout.write(self.style.ERROR(e.message))



# -*- coding: utf-8 -*-
from django.db import migrations, models
import waldur_auth_valimo.models
import django.utils.timezone
from django.conf import settings
import django_fsm
import waldur_core.core.fields
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthResult',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('phone', models.CharField(max_length=30)),
                ('message', models.CharField(default=waldur_auth_valimo.models._default_message, help_text='This message will be shown to user.', max_length=4)),
                ('state', django_fsm.FSMField(default='Scheduled', max_length=50, choices=[('Scheduled', 'Scheduled'), ('Processing', 'Processing'), ('OK', 'OK'), ('Canceled', 'Canceled'), ('Erred', 'Erred')])),
                ('details', models.CharField(help_text='Cancellation details.', max_length=255, blank=True)),
                ('backend_transaction_id', models.CharField(max_length=100, blank=True)),
                ('user', models.ForeignKey(related_name='auth_valimo_results', to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.db.models import OuterRef, Subquery
from rest_framework.filters import BaseFilterBackend

from waldur_core.core import filters as core_filters
from waldur_core.core.utils import order_with_nulls, get_ordering
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import models as structure_models
from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import utils as invoice_utils

from . import models


class PriceEstimateScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return models.PriceEstimate.get_estimated_models()

    def get_field_name(self):
        return 'scope'


class CustomerEstimatedCostFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):

        order_by = get_ordering(request)
        if order_by not in ('estimated_cost', '-estimated_cost'):
            return queryset

        ct = ContentType.objects.get_for_model(structure_models.Customer)
        estimates = models.PriceEstimate.objects.filter(content_type=ct, object_id=OuterRef('pk'))
        queryset = queryset.annotate(estimated_cost=Subquery(estimates.values('total')[:1]))
        return order_with_nulls(queryset, order_by)


class CustomerCurrentCostFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):

        order_by = get_ordering(request)
        if order_by not in ('current_cost', '-current_cost'):
            return queryset

        year, month = invoice_utils.parse_period(request.query_params)
        invoices = invoice_models.Invoice.objects.filter(year=year, month=month, customer=OuterRef('pk'))
        queryset = queryset.annotate(current_cost=Subquery(invoices.values('current_cost')[:1]))
        return order_with_nulls(queryset, order_by)


structure_filters.ExternalCustomerFilterBackend.register(CustomerEstimatedCostFilter())
structure_filters.ExternalCustomerFilterBackend.register(CustomerCurrentCostFilter())



from waldur_core.core import tasks as core_tasks

from . import models


class TenantCreateErrorTask(core_tasks.ErrorStateTransitionTask):

    def execute(self, tenant):
        super(TenantCreateErrorTask, self).execute(tenant)
        # Delete network and subnet if they were not created on backend,
        # mark as erred if they were created
        network = tenant.networks.first()
        subnet = network.subnets.first()
        if subnet.state == models.SubNet.States.CREATION_SCHEDULED:
            subnet.delete()
        else:
            super(TenantCreateErrorTask, self).execute(subnet)
        if network.state == models.Network.States.CREATION_SCHEDULED:
            network.delete()
        else:
            super(TenantCreateErrorTask, self).execute(network)


class TenantCreateSuccessTask(core_tasks.StateTransitionTask):

    def execute(self, tenant):
        network = tenant.networks.first()
        subnet = network.subnets.first()
        self.state_transition(network, 'set_ok')
        self.state_transition(subnet, 'set_ok')
        self.state_transition(tenant, 'set_ok')
        return super(TenantCreateSuccessTask, self).execute(tenant)


class TenantPullQuotas(core_tasks.BackgroundTask):
    name = 'openstack.TenantPullQuotas'

    def is_equal(self, other_task):
        return self.name == other_task.get('name')

    def run(self):
        from . import executors
        for tenant in models.Tenant.objects.filter(state=models.Tenant.States.OK):
            executors.TenantPullQuotasExecutor.execute(tenant)



from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from python_freeipa import exceptions as freeipa_exceptions
from rest_framework import decorators, exceptions, response, status

from waldur_core.core import views as core_views
from . import backend, filters, models, serializers, tasks


class CheckExtensionMixin(core_views.CheckExtensionMixin):
    extension_name = 'WALDUR_FREEIPA'


class ProfileViewSet(CheckExtensionMixin, core_views.ActionsViewSet):
    queryset = models.Profile.objects.all()
    filterset_class = filters.ProfileFilter
    serializer_class = serializers.ProfileSerializer
    disabled_actions = ['destroy']
    lookup_field = 'uuid'

    def get_queryset(self):
        qs = super(ProfileViewSet, self).get_queryset()
        if self.request.user.is_staff:
            return qs
        return qs.filter(user=self.request.user)

    @transaction.atomic()
    def perform_create(self, serializer):
        profile = serializer.save()
        try:
            backend.FreeIPABackend().create_profile(profile)
            tasks.schedule_sync()
        except freeipa_exceptions.DuplicateEntry:
            raise exceptions.ValidationError({
                'username': _('Profile with such name already exists.')
            })

    @decorators.action(detail=True, methods=['post'])
    def update_ssh_keys(self, request, uuid=None):
        profile = self.get_object()
        try:
            backend.FreeIPABackend().update_ssh_keys(profile)
        except freeipa_exceptions.NotFound:
            profile.delete()
            return response.Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return response.Response(status=status.HTTP_200_OK)

    @decorators.action(detail=True, methods=['post'])
    @transaction.atomic()
    def disable(self, request, uuid=None):
        profile = self.get_object()
        if not profile.is_active:
            raise exceptions.ValidationError({
                'detail': _('Profile is already disabled.')
            })
        try:
            backend.FreeIPABackend().disable_profile(profile)
        except freeipa_exceptions.NotFound:
            profile.delete()
            return response.Response(status=status.HTTP_204_NO_CONTENT)
        except freeipa_exceptions.AlreadyInactive:
            profile.is_active = False
            profile.save(update_fields=['is_active'])
            return response.Response(status=status.HTTP_200_OK)
        else:
            profile.is_active = False
            profile.save(update_fields=['is_active'])
            return response.Response(status=status.HTTP_200_OK)

    @decorators.action(detail=True, methods=['post'])
    @transaction.atomic()
    def enable(self, request, uuid=None):
        profile = self.get_object()
        if profile.is_active:
            raise exceptions.ValidationError({
                'detail': _('Profile is already enabled.')
            })
        try:
            backend.FreeIPABackend().enable_profile(profile)
        except freeipa_exceptions.NotFound:
            profile.delete()
            return response.Response(status=status.HTTP_204_NO_CONTENT)
        except freeipa_exceptions.AlreadyActive:
            profile.is_active = True
            profile.save(update_fields=['is_active'])
            return response.Response(status=status.HTTP_200_OK)
        else:
            profile.is_active = True
            profile.save(update_fields=['is_active'])
            return response.Response(status=status.HTTP_200_OK)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-08-04 12:54
import waldur_ansible.playbook_jobs.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('playbook_jobs', '0002_add_parameter_order'),
    ]

    operations = [
        migrations.AddField(
            model_name='playbook',
            name='image',
            field=models.ImageField(blank=True, null=True, upload_to=waldur_ansible.playbook_jobs.models.get_upload_path),
        ),
    ]



import functools
from io import BytesIO
import logging

from django.conf import settings
from django.db import transaction, IntegrityError
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.functional import cached_property
from jira.client import _get_template_list
from jira.utils import json_loads
from rest_framework import status

from waldur_core.core.models import StateMixin
from waldur_core.structure import ServiceBackend, ServiceBackendError
from waldur_core.structure.utils import update_pulled_fields

from .jira_fix import JIRA, JIRAError
from . import models

logger = logging.getLogger(__name__)


class JiraBackendError(ServiceBackendError):
    pass


def check_captcha(e):
    if e.response is None:
        return False
    if not hasattr(e.response, 'headers'):
        return False
    if 'X-Seraph-LoginReason' not in e.response.headers:
        return False
    return e.response.headers['X-Seraph-LoginReason'] == 'AUTHENTICATED_FAILED'


def reraise_exceptions(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except JIRAError as e:
            raise JiraBackendError(e)

    return wrapped


class JiraBackend(ServiceBackend):
    """ Waldur interface to JIRA.
        http://pythonhosted.org/jira/
        http://docs.atlassian.com/jira/REST/latest/
    This class can be overridden in other modules that use JIra.
    Declare used models of other modules:
        - model_comment
        - model_issue
        - model_attachment
    And if necessary, overridden methods
        - _backend_issue_to_issue
        - _backend_comment_to_comment
        - _backend_attachment_to_attachment

    """

    model_comment = models.Comment
    model_issue = models.Issue
    model_attachment = models.Attachment
    model_project = models.Project

    def __init__(self, settings, project=None, verify=False):
        self.settings = settings
        self.project = project
        self.verify = verify

    def pull_service_properties(self):
        self.pull_project_templates()
        self.pull_priorities()

    def ping(self, raise_exception=False):
        try:
            self.manager.myself()
        except JIRAError as e:
            if raise_exception:
                raise JiraBackendError(e)
            return False
        else:
            return True

    @reraise_exceptions
    def get_resources_for_import(self):
        return [{
            'name': proj.name,
            'backend_id': proj.key,
        } for proj in self.manager.projects()]

    @staticmethod
    def convert_field(value, choices, mapping=None):
        """ Reverse mapping for choice fields """
        if mapping:
            mapping = {v: k for k, v in mapping.items()}
            value = mapping.get(value, value)

        try:
            return next(k for k, v in choices if v == value)
        except StopIteration:
            return 0

    @cached_property
    def manager(self):
        if self.settings.token:
            if getattr(self.settings, 'email', None):
                basic_auth = (self.settings.email, self.settings.token)
            else:
                basic_auth = (self.settings.username, self.settings.token)
        else:
            basic_auth = (self.settings.username, self.settings.password)

        try:
            return JIRA(
                server=self.settings.backend_url,
                options={'verify': self.verify},
                basic_auth=basic_auth,
                validate=False)
        except JIRAError as e:
            if check_captcha(e):
                raise JiraBackendError('JIRA CAPTCHA is triggered. Please reset credentials.')
            raise JiraBackendError(e)

    @reraise_exceptions
    def get_field_id_by_name(self, field_name):
        if not field_name:
            return None
        try:
            fields = getattr(self, '_fields')
        except AttributeError:
            fields = self._fields = self.manager.fields()
        try:
            return next(f['id'] for f in fields if field_name in f['clauseNames'])
        except StopIteration:
            raise JiraBackendError("Can't find custom field %s" % field_name)

    @reraise_exceptions
    def get_project_templates(self):
        url = self.manager._options['server'] + '/rest/project-templates/latest/templates'

        response = self.manager._session.get(url)
        json_data = json_loads(response)
        return _get_template_list(json_data)

    @reraise_exceptions
    def pull_project_templates(self):
        backend_templates = self.get_project_templates()
        with transaction.atomic():
            for template in backend_templates:
                backend_id = template['projectTemplateModuleCompleteKey']
                icon_url = self.manager._options['server'] + template['iconUrl']
                models.ProjectTemplate.objects.update_or_create(
                    backend_id=backend_id,
                    defaults={
                        'name': template['name'],
                        'description': template['description'],
                        'icon_url': icon_url,
                    })

    @reraise_exceptions
    def pull_priorities(self):
        backend_priorities = self.manager.priorities()
        with transaction.atomic():
            backend_priorities_map = {
                priority.id: priority for priority in backend_priorities
            }

            waldur_priorities = {
                priority.backend_id: priority
                for priority in models.Priority.objects.filter(settings=self.settings)
            }

            stale_priorities = set(waldur_priorities.keys()) - set(backend_priorities_map.keys())
            models.Priority.objects.filter(backend_id__in=stale_priorities)

            for priority in backend_priorities:
                models.Priority.objects.update_or_create(
                    backend_id=priority.id,
                    settings=self.settings,
                    defaults={
                        'name': priority.name,
                        'description': priority.description,
                        'icon_url': priority.iconUrl,
                    })

    @reraise_exceptions
    def import_priority(self, priority):
        return models.Priority(
            backend_id=priority.id,
            settings=self.settings,
            name=priority.name,
            description=getattr(property, 'description', ''),
            icon_url=priority.iconUrl,
        )

    @reraise_exceptions
    def get_project(self, project_id):
        return self.manager.project(project_id)

    @cached_property
    def default_assignee(self):
        # JIRA REST API basic authentication accepts either username or email.
        # But create project endpoint does not accept email.
        # Therefore we need to get username for the logged in user.
        user = self.manager.myself()
        return user['name']

    @reraise_exceptions
    def create_project(self, project):
        self.manager.create_project(
            key=project.backend_id,
            name=project.name,
            assignee=self.default_assignee,
            template_name=project.template.name,
        )
        self.pull_issue_types(project)

    def pull_issue_types(self, project):
        backend_project = self.get_project(project.backend_id)
        backend_issue_types = {
            issue_type.id: issue_type
            for issue_type in backend_project.issueTypes
        }
        project_issue_types = {
            issue_type.backend_id: issue_type
            for issue_type in project.issue_types.all()
        }
        global_issue_types = {
            issue_type.backend_id: issue_type
            for issue_type in models.IssueType.objects.filter(settings=self.settings)
        }

        new_issue_types = set(backend_issue_types.keys()) - set(project_issue_types.keys())
        for issue_type_id in new_issue_types:
            if issue_type_id in global_issue_types:
                issue_type = global_issue_types[issue_type_id]
            else:
                issue_type = self.import_issue_type(backend_issue_types[issue_type_id])
                issue_type.save()
            project.issue_types.add(issue_type)

        stale_issue_types = set(project_issue_types.keys()) - set(backend_issue_types.keys())
        project.issue_types.filter(backend_id__in=stale_issue_types).delete()

        common_issue_types = set(project_issue_types.keys()) & set(backend_issue_types.keys())
        for issue_type_id in common_issue_types:
            issue_type = project_issue_types[issue_type_id]
            imported_issue_type = self.import_issue_type(backend_issue_types[issue_type_id])
            update_pulled_fields(issue_type, imported_issue_type, (
                'name', 'description', 'icon_url', 'subtask'
            ))

    def import_issue_type(self, backend_issue_type):
        return models.IssueType(
            settings=self.settings,
            backend_id=backend_issue_type.id,
            name=backend_issue_type.name,
            description=backend_issue_type.description,
            icon_url=backend_issue_type.iconUrl,
            subtask=backend_issue_type.subtask,
        )

    @reraise_exceptions
    def update_project(self, project):
        backend_project = self.manager.project(project.backend_id)
        backend_project.update(name=project.name)

    @reraise_exceptions
    def delete_project(self, project):
        self.manager.delete_project(project.backend_id)

    @reraise_exceptions
    def create_issue(self, issue):
        args = self._issue_to_dict(issue)
        backend_issue = self.manager.create_issue(**args)
        self._backend_issue_to_issue(backend_issue, issue)
        issue.save()

    def create_issue_from_jira(self, project, key):
        backend_issue = self.get_backend_issue(key)
        if not backend_issue:
            logger.debug('Unable to create issue with key=%s, '
                         'because it has already been deleted on backend.', key)
            return

        issue = self.model_issue(project=project, backend_id=key, state=StateMixin.States.OK)
        self._backend_issue_to_issue(backend_issue, issue)
        try:
            issue.save()
        except IntegrityError:
            logger.debug('Unable to create issue with key=%s, '
                         'because it has been created in another thread.', key)

    def update_issue(self, issue):
        backend_issue = self.get_backend_issue(issue.backend_id)
        if not backend_issue:
            logger.debug('Unable to update issue with key=%s, '
                         'because it has already been deleted on backend.', issue.backend_id)
            return

        backend_issue.update(summary=issue.summary, description=issue.get_description())

    def update_issue_from_jira(self, issue):
        start_time = timezone.now()

        backend_issue = self.get_backend_issue(issue.backend_id)
        if not backend_issue:
            logger.debug('Unable to update issue with key=%s, '
                         'because it has already been deleted on backend.', issue.backend_id)
            return

        issue.refresh_from_db()

        if issue.modified > start_time:
            logger.debug('Skipping issue update with key=%s, '
                         'because it has been updated from other thread.', issue.backend_id)
            return

        self._backend_issue_to_issue(backend_issue, issue)
        issue.save()

    def delete_issue(self, issue):
        backend_issue = self.get_backend_issue(issue.backend_id)
        if backend_issue:
            backend_issue.delete()
        else:
            logger.debug('Unable to delete issue with key=%s, '
                         'because it has already been deleted on backend.', issue.backend_id)

    def delete_issue_from_jira(self, issue):
        backend_issue = self.get_backend_issue(issue.backend_id)
        if not backend_issue:
            issue.delete()
        else:
            logger.debug('Skipping issue deletion with key=%s, '
                         'because it still exists on backend.', issue.backend_id)

    @reraise_exceptions
    def create_comment(self, comment):
        backend_comment = self.manager.add_comment(comment.issue.backend_id, comment.prepare_message())
        comment.backend_id = backend_comment.id
        comment.save(update_fields=['backend_id'])

    def create_comment_from_jira(self, issue, comment_backend_id):
        backend_comment = self.get_backend_comment(issue.backend_id, comment_backend_id)
        if not backend_comment:
            logger.debug('Unable to create comment with id=%s, '
                         'because it has already been deleted on backend.', comment_backend_id)
            return

        comment = self.model_comment(issue=issue, backend_id=comment_backend_id, state=StateMixin.States.OK)
        self._backend_comment_to_comment(backend_comment, comment)

        try:
            comment.save()
        except IntegrityError:
            logger.debug('Unable to create comment issue_id=%s, backend_id=%s, '
                         'because it already exists  n Waldur.', issue.id, comment_backend_id)

    def update_comment(self, comment):
        backend_comment = self.get_backend_comment(comment.issue.backend_id, comment.backend_id)
        if not backend_comment:
            logger.debug('Unable to update comment with id=%s, '
                         'because it has already been deleted on backend.', comment.id)
            return

        backend_comment.update(body=comment.prepare_message())

    def update_comment_from_jira(self, comment):
        backend_comment = self.get_backend_comment(comment.issue.backend_id, comment.backend_id)
        if not backend_comment:
            logger.debug('Unable to update comment with id=%s, '
                         'because it has already been deleted on backend.', comment.id)
            return

        comment.state = StateMixin.States.OK
        self._backend_comment_to_comment(backend_comment, comment)
        comment.save()

    @reraise_exceptions
    def delete_comment(self, comment):
        backend_comment = self.get_backend_comment(comment.issue.backend_id, comment.backend_id)
        if backend_comment:
            backend_comment.delete()
        else:
            logger.debug('Unable to delete comment with id=%s, '
                         'because it has already been deleted on backend.', comment.id)

    def delete_comment_from_jira(self, comment):
        backend_comment = self.get_backend_comment(comment.issue.backend_id, comment.backend_id)
        if not backend_comment:
            comment.delete()
        else:
            logger.debug('Skipping comment deletion with id=%s, '
                         'because it still exists on backend.', comment.id)

    @reraise_exceptions
    def create_attachment(self, attachment):
        backend_issue = self.get_backend_issue(attachment.issue.backend_id)
        if not backend_issue:
            logger.debug('Unable to add attachment to issue with id=%s, '
                         'because it has already been deleted on backend.', attachment.issue.id)
            return

        backend_attachment = self.manager.waldur_add_attachment(backend_issue, attachment.file.path)
        attachment.backend_id = backend_attachment.id
        attachment.save(update_fields=['backend_id'])

    @reraise_exceptions
    def delete_attachment(self, attachment):
        backend_attachment = self.get_backend_attachment(attachment.backend_id)
        if backend_attachment:
            backend_attachment.delete()
        else:
            logger.debug('Unable to remove attachment with id=%s, '
                         'because it has already been deleted on backend.', attachment.id)

    @reraise_exceptions
    def import_project_issues(self, project, start_at=0, max_results=50, order=None):
        waldur_issues = list(self.model_issue.objects.filter(project=project, backend_id__isnull=False).
                             values_list('backend_id', flat=True))

        jql = 'project=%s' % project.backend_id
        if order:
            jql += ' ORDER BY %s' % order

        for backend_issue in self.manager.search_issues(jql, startAt=start_at, maxResults=max_results, fields='*all'):
            key = backend_issue.key
            if key in waldur_issues:
                logger.debug('Skipping import of issue with key=%s, '
                             'because it already exists in Waldur.', key)
                continue

            issue = self.model_issue(project=project, backend_id=key, state=StateMixin.States.OK)
            self._backend_issue_to_issue(backend_issue, issue)
            issue.save()

            attachment_synchronizer = AttachmentSynchronizer(self, issue, backend_issue)
            attachment_synchronizer.perform_update()

            for backend_comment in backend_issue.fields.comment.comments:
                tmp = issue.comments.model()
                tmp.clean_message(backend_comment.body)
                issue.comments.create(
                    user=tmp.user,
                    message=tmp.message,
                    created=parse_datetime(backend_comment.created),
                    backend_id=backend_comment.id,
                    state=issue.comments.model.States.OK)

    def _import_project(self, project_backend_id, service_project_link, state):
        backend_project = self.get_project(project_backend_id)
        project = self.model_project(
            service_project_link=service_project_link,
            backend_id=project_backend_id,
            state=state)
        self._backend_project_to_project(backend_project, project)
        project.save()
        return project

    def import_project(self, project_backend_id, service_project_link):
        project = self._import_project(project_backend_id, service_project_link,
                                       models.Project.States.OK)
        self.import_project_issues(project)
        return project

    def import_project_scheduled(self, project_backend_id, service_project_link):
        project = self._import_project(project_backend_id, service_project_link,
                                       models.Project.States.OK)
        return project

    def import_project_batch(self, project):
        max_results = settings.WALDUR_JIRA.get('ISSUE_IMPORT_LIMIT')
        start_at = project.action_details.get('current_issue', 0)
        self.import_project_issues(project, order='id', start_at=start_at, max_results=max_results)

        total_added = project.action_details.get('current_issue', 0) + max_results

        if total_added >= project.action_details.get('issues_count', 0):
            project.action_details['current_issue'] = project.action_details.get('issues_count', 0)
            project.action_details['percentage'] = 100
            project.runtime_state = 'success'
        else:
            project.action_details['current_issue'] = total_added
            project.action_details['percentage'] = int((project.action_details['current_issue'] /
                                                        project.action_details['issues_count']) * 100)

        project.save()
        return max_results

    def get_backend_comment(self, issue_backend_id, comment_backend_id):
        return self._get_backend_obj('comment')(issue_backend_id, comment_backend_id)

    def get_backend_issue(self, issue_backend_id):
        return self._get_backend_obj('issue')(issue_backend_id)

    def get_backend_attachment(self, attachment_backend_id):
        return self._get_backend_obj('attachment')(attachment_backend_id)

    def update_attachment_from_jira(self, issue):
        backend_issue = self.get_backend_issue(issue.backend_id)
        AttachmentSynchronizer(self, issue, backend_issue).perform_update()

    def delete_old_comments(self, issue):
        backend_issue = self.get_backend_issue(issue.backend_id)
        CommentSynchronizer(self, issue, backend_issue).perform_update()

    @reraise_exceptions
    def _get_backend_obj(self, method):
        def f(*args, **kwargs):
            try:
                func = getattr(self.manager, method)
                backend_obj = func(*args, **kwargs)
            except JIRAError as e:
                if e.status_code == status.HTTP_404_NOT_FOUND:
                    logger.debug('Jira object {} has been already deleted on backend'.format(method))
                    return
                else:
                    raise e
            return backend_obj
        return f

    def _backend_issue_to_issue(self, backend_issue, issue):
        priority = self._get_or_create_priority(issue.project, backend_issue.fields.priority)
        issue_type = self._get_or_create_issue_type(issue.project, backend_issue.fields.issuetype)
        resolution_sla = self._get_resolution_sla(backend_issue)

        for obj in ['assignee', 'creator', 'reporter']:
            backend_obj = getattr(backend_issue.fields, obj, None)
            fields = [
                ['name', 'displayName'],
                ['username', 'name'],
                ['email', 'emailAddress'],
            ]

            for waldur_key, backend_key in fields:
                value = getattr(backend_obj, backend_key, '')
                setattr(issue, obj + '_' + waldur_key, value)

        issue.priority = priority
        issue.summary = backend_issue.fields.summary
        issue.description = backend_issue.fields.description or ''
        issue.type = issue_type
        issue.status = backend_issue.fields.status.name or ''
        issue.resolution = (backend_issue.fields.resolution and backend_issue.fields.resolution.name) or ''
        issue.resolution_date = backend_issue.fields.resolutiondate
        issue.resolution_sla = resolution_sla
        issue.backend_id = backend_issue.key

    def _backend_comment_to_comment(self, backend_comment, comment):
        comment.update_message(backend_comment.body)

    def _backend_attachment_to_attachment(self, backend_attachment, attachment):
        pass

    def _backend_project_to_project(self, backend_project, project):
        project.name = backend_project.name
        project.description = backend_project.description

    def _get_or_create_priority(self, project, backend_priority):
        try:
            priority = models.Priority.objects.get(
                settings=project.service_project_link.service.settings,
                backend_id=backend_priority.id
            )
        except models.Priority.DoesNotExist:
            priority = self.import_priority(backend_priority)
            priority.save()
        return priority

    def _get_or_create_issue_type(self, project, backend_issue_type):
        try:
            issue_type = models.IssueType.objects.get(
                settings=project.service_project_link.service.settings,
                backend_id=backend_issue_type.id
            )
        except models.IssueType.DoesNotExist:
            issue_type = self.import_issue_type(backend_issue_type)
            issue_type.save()
            project.issue_types.add(issue_type)
        return issue_type

    def _get_resolution_sla(self, backend_issue):
        issue_settings = settings.WALDUR_JIRA.get('ISSUE')
        field_name = self.get_field_id_by_name(issue_settings['resolution_sla_field'])
        value = getattr(backend_issue.fields, field_name, None)

        if value and hasattr(value, 'ongoingCycle'):
            milliseconds = value.ongoingCycle.remainingTime.millis
            if milliseconds:
                resolution_sla = milliseconds / 1000
        else:
            resolution_sla = None
        return resolution_sla

    def _issue_to_dict(self, issue):
        args = dict(
            project=issue.project.backend_id,
            summary=issue.summary,
            description=issue.get_description(),
            issuetype={'name': issue.type.name},
        )

        if issue.priority:
            args['priority'] = {'name': issue.priority.name}

        if issue.parent:
            args['parent'] = {'key': issue.parent.backend_id}

        return args

    def _get_property(self, object_name, object_id, property_name):
        url = self.manager._get_url('{0}/{1}/properties/{2}'.format(object_name, object_id, property_name))
        response = self.manager._session.get(url)
        return response.json()

    def get_issues_count(self, project_key):
        base = '{server}/rest/{rest_path}/{rest_api_version}/{path}'
        page_params = {'jql': 'project=%s' % project_key,
                       'validateQuery': True,
                       'startAt': 0,
                       'fields': [],
                       'maxResults': 0,
                       'expand': None}
        result = self.manager._get_json('search', params=page_params, base=base)
        return result['total']


class AttachmentSynchronizer:
    def __init__(self, backend, current_issue, backend_issue):
        self.backend = backend
        self.current_issue = current_issue
        self.backend_issue = backend_issue

    def perform_update(self):
        if self.stale_attachment_ids:
            self.backend.model_attachment.objects.filter(backend_id__in=self.stale_attachment_ids).delete()

        for attachment_id in self.new_attachment_ids:
            self._add_attachment(
                self.current_issue,
                self.get_backend_attachment(attachment_id)
            )

        for attachment_id in self.updated_attachments_ids:
            self._update_attachment(
                self.current_issue,
                self.get_backend_attachment(attachment_id),
                self.get_current_attachment(attachment_id)
            )

    def get_current_attachment(self, attachment_id):
        return self.current_attachments_map[attachment_id]

    def get_backend_attachment(self, attachment_id):
        return self.backend_attachments_map[attachment_id]

    @cached_property
    def current_attachments_map(self):
        return {
            str(attachment.backend_id): attachment
            for attachment in self.current_issue.attachments.all()
        }

    @cached_property
    def current_attachments_ids(self):
        return set(self.current_attachments_map.keys())

    @cached_property
    def backend_attachments_map(self):
        return {
            str(attachment.id): attachment
            for attachment in self.backend_issue.fields.attachment
        }

    @cached_property
    def backend_attachments_ids(self):
        return set(self.backend_attachments_map.keys())

    @cached_property
    def stale_attachment_ids(self):
        return self.current_attachments_ids - self.backend_attachments_ids

    @cached_property
    def new_attachment_ids(self):
        return self.backend_attachments_ids - self.current_attachments_ids

    @cached_property
    def updated_attachments_ids(self):
        return filter(self._is_attachment_updated, self.backend_attachments_ids)

    def _is_attachment_updated(self, attachment_id):
        """
        Attachment is considered updated if its thumbnail just has been created.
        """

        if not getattr(self.get_backend_attachment(attachment_id), 'thumbnail', False):
            return False

        if attachment_id not in self.current_attachments_ids:
            return False

        if self.get_current_attachment(attachment_id).thumbnail:
            return False

        return True

    def _download_file(self, url):
        """
        Download file from URL using secure JIRA session.
        :return: byte stream
        :raises: requests.RequestException
        """
        session = self.backend.manager._session
        response = session.get(url)
        response.raise_for_status()
        return BytesIO(response.content)

    def _add_attachment(self, issue, backend_attachment):
        attachment = self.backend.model_attachment(issue=issue,
                                                   backend_id=backend_attachment.id,
                                                   state=StateMixin.States.OK)
        thumbnail = getattr(backend_attachment, 'thumbnail', False) and getattr(attachment, 'thumbnail', False)

        try:
            content = self._download_file(backend_attachment.content)
            if thumbnail:
                thumbnail_content = self._download_file(backend_attachment.thumbnail)

        except JIRAError as error:
            logger.error('Unable to load attachment for issue with backend id {backend_id}. Error: {error}).'
                         .format(backend_id=issue.backend_id, error=error))
            return

        self.backend._backend_attachment_to_attachment(backend_attachment, attachment)

        try:
            attachment.save()
        except IntegrityError:
            logger.debug('Unable to create attachment issue_id=%s, backend_id=%s, '
                         'because it already exists in Waldur.', issue.id, backend_attachment.id)

        attachment.file.save(backend_attachment.filename, content, save=True)

        if thumbnail:
            attachment.thumbnail.save(backend_attachment.filename, thumbnail_content, save=True)

    def _update_attachment(self, issue, backend_attachment, current_attachment):
        try:
            content = self._download_file(backend_attachment.thumbnail)
        except JIRAError as error:
            logger.error('Unable to load attachment thumbnail for issue with backend id {backend_id}. Error: {error}).'
                         .format(backend_id=issue.backend_id, error=error))
            return

        current_attachment.thumbnail.save(backend_attachment.filename, content, save=True)


class CommentSynchronizer:
    def __init__(self, backend, current_issue, backend_issue):
        self.backend = backend
        self.current_issue = current_issue
        self.backend_issue = backend_issue

    def perform_update(self):
        if self.stale_comments_ids:
            self.backend.model_comment.objects.filter(backend_id__in=self.stale_comments_ids).delete()

    def get_current_comment(self, comment_id):
        return self.current_comments_map[comment_id]

    def get_backend_comment(self, comment_id):
        return self.backend_comments_map[comment_id]

    @cached_property
    def current_comments_map(self):
        return {
            str(comment.backend_id): comment
            for comment in self.current_issue.comments.all()
        }

    @cached_property
    def current_comments_ids(self):
        return set(self.current_comments_map.keys())

    @cached_property
    def backend_comments_map(self):
        return {
            str(comment.id): comment
            for comment in self.backend_issue.fields.comment.comments
        }

    @cached_property
    def backend_comments_ids(self):
        return set(self.backend_comments_map.keys())

    @cached_property
    def stale_comments_ids(self):
        return self.current_comments_ids - self.backend_comments_ids



import collections
import logging

import requests
from django.conf import settings

from . import exceptions

logger = logging.getLogger(__name__)


Coordinates = collections.namedtuple('Coordinates', ('latitude', 'longitude'))


def get_coordinates_by_ip(ip_address):
    """
    Return coordinates by IP or hostname.
    :param ip_address: IP or hostname
    """
    if not settings.IPSTACK_ACCESS_KEY:
        raise exceptions.GeoIpException("IPSTACK_ACCESS_KEY is empty.")

    url = 'http://api.ipstack.com/{}?access_key={}&output=json&legacy=1'.format(
        ip_address,
        settings.IPSTACK_ACCESS_KEY)  # We don't use https, because current plan does not support HTTPS Encryption

    try:
        response = requests.get(url)
    except requests.exceptions.RequestException as e:
        raise exceptions.GeoIpException("Request to geoip API %s failed: %s" % (url, e))

    if response.ok:
        data = response.json()
        latitude = data.get('latitude')
        longitude = data.get('longitude')

        if latitude and longitude:
            return Coordinates(latitude=latitude,
                               longitude=longitude)

    params = (url, response.status_code, response.text)
    raise exceptions.GeoIpException("Request to geoip API %s failed: %s %s" % params)


def detect_coordinates(instance):
    try:
        coordinates = instance.detect_coordinates()
    except exceptions.GeoIpException as e:
        logger.warning('Unable to detect coordinates for %s: %s.', instance, e)
        return

    if coordinates:
        instance.latitude = coordinates.latitude
        instance.longitude = coordinates.longitude
        instance.save(update_fields=['latitude', 'longitude'])



from rest_framework import serializers

from . import models


class ChecklistSerializer(serializers.ModelSerializer):
    questions_count = serializers.ReadOnlyField(source='questions.count')

    class Meta:
        model = models.Checklist
        fields = ('uuid', 'name', 'description', 'questions_count')


class QuestionSerializer(serializers.ModelSerializer):
    category_uuid = serializers.ReadOnlyField(source='category.uuid')

    class Meta:
        model = models.Question
        fields = ('uuid', 'description', 'solution', 'category_uuid')


class AnswerListSerializer(serializers.ModelSerializer):
    question_uuid = serializers.ReadOnlyField(source='question.uuid')

    class Meta:
        model = models.Answer
        fields = ('question_uuid', 'value')


class AnswerSubmitSerializer(serializers.Serializer):
    question_uuid = serializers.UUIDField()
    value = serializers.NullBooleanField()



from django.utils.translation import ugettext_lazy as _
from rest_framework import status
from rest_framework.exceptions import ValidationError, PermissionDenied
from rest_framework.generics import get_object_or_404
from rest_framework.mixins import ListModelMixin, CreateModelMixin
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet

from waldur_core.structure.models import Customer, Project
from waldur_core.structure.permissions import is_administrator

from . import models, serializers


class ChecklistView(ListModelMixin, GenericViewSet):
    queryset = models.Checklist.objects.all()
    serializer_class = serializers.ChecklistSerializer


class QuestionsView(ListModelMixin, GenericViewSet):
    serializer_class = serializers.QuestionSerializer

    def get_queryset(self):
        return models.Question.objects.filter(checklist__uuid=self.kwargs['checklist_uuid'])


class StatsView(APIView):
    def get(self, request, checklist_uuid, format=None):
        if not request.user.is_staff and not request.user.is_support:
            raise PermissionDenied()

        checklist = get_object_or_404(models.Checklist, uuid=checklist_uuid)
        total_questions = checklist.questions.count()
        points = []
        for customer in Customer.objects.all():
            projects_count = customer.projects.count()
            positive_count = models.Answer.objects.filter(
                project__in=customer.projects.all(),
                question__checklist=checklist,
                value=True,
            ).count()
            points.append(dict(
                name=customer.name,
                uuid=customer.uuid,
                latitude=customer.latitude,
                longitude=customer.longitude,
                score=round(100 * positive_count / max(1, projects_count * total_questions), 2)
            ))
        return Response(points)


class ProjectStatsView(APIView):
    def get(self, request, project_uuid, format=None):
        try:
            project = Project.objects.get(uuid=project_uuid)
        except Project.DoesNotExist:
            raise ValidationError(_('Project does not exist.'))

        is_administrator(request, self, project)

        checklists = []
        for checklist in models.Checklist.objects.all():
            qs = models.Answer.objects.filter(
                project=project,
                question__checklist=checklist
            )
            total = checklist.questions.count()
            positive_count = qs.filter(value=True).count()
            negative_count = qs.filter(value=False).count()
            unknown_count = total - positive_count - negative_count
            checklists.append(dict(
                name=checklist.name,
                uuid=checklist.uuid,
                positive_count=positive_count,
                negative_count=negative_count,
                unknown_count=unknown_count,
                score=round(100 * positive_count / total, 2) if total > 0 else 100,  # consider empty lists as fully compliant
            ))
        return Response(checklists)


class AnswersListView(ListModelMixin, GenericViewSet):
    serializer_class = serializers.AnswerListSerializer

    def get_queryset(self):
        return models.Answer.objects.filter(
            question__checklist__uuid=self.kwargs['checklist_uuid'],
            project__uuid=self.kwargs['project_uuid'],
        )


class AnswersSubmitView(CreateModelMixin, GenericViewSet):
    serializer_class = serializers.AnswerSubmitSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)

        try:
            project = Project.objects.get(uuid=self.kwargs['project_uuid'])
        except Project.DoesNotExist:
            raise ValidationError(_('Project does not exist.'))

        is_administrator(request, self, project)

        try:
            checklist = models.Checklist.objects.get(uuid=self.kwargs['checklist_uuid'])
        except models.Checklist.DoesNotExist:
            raise ValidationError(_('Checklist does not exist.'))

        for answer in serializer.validated_data:
            try:
                question = checklist.questions.get(uuid=answer['question_uuid'])
            except models.Question.DoesNotExist:
                raise ValidationError(_('Question does not exist.'))

            models.Answer.objects.update_or_create(
                question=question,
                project=project,
                defaults={
                    'user': request.user,
                    'value': answer['value'],
                }
            )

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)



from celery import chain

from waldur_core.core import executors
from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.structure import executors as structure_executors

from . import models
from .tasks import SetInstanceErredTask


class VolumeCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume, 'create_volume', state_transition='begin_creating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30)
        )


class VolumeDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_volume, 'delete_volume', state_transition='begin_deleting')
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_deleting')


class VolumeDetachExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume, 'detach_volume', state_transition='begin_updating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error'
            ).set(countdown=10)
        )


class VolumeAttachExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume, 'attach_volume', state_transition='begin_updating'),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='inuse',
                erred_state='error',
            ).set(countdown=10)
        )


class InstanceCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, image=None, size=None, ssh_key=None, volume=None):
        kwargs = {
            'backend_image_id': image.backend_id,
            'backend_size_id': size.backend_id
        }
        if ssh_key is not None:
            kwargs['ssh_key_uuid'] = ssh_key.uuid.hex

        serialized_volume = core_utils.serialize_instance(volume)

        return chain(
            core_tasks.StateTransitionTask().si(
                serialized_volume,
                state_transition='begin_creating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='create_instance',
                state_transition='begin_creating',
                **kwargs),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='running',
                erred_state='error'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                backend_method='pull_instance_volume',
                success_runtime_state='inuse',
            ),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_public_ips'),
        )

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        serialized_volume = core_utils.serialize_instance(instance.volume_set.first())
        return chain(
            core_tasks.StateTransitionTask().si(serialized_volume, state_transition='set_ok'),
            core_tasks.StateTransitionTask().si(serialized_instance, state_transition='set_ok')
        )

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return SetInstanceErredTask().s(serialized_instance)


class InstanceResizeExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        size = kwargs.pop('size')
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                backend_method='resize_instance',
                state_transition='begin_updating',
                size_id=size.backend_id
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='stopped',
                erred_state='error'
            ).set(countdown=30)
        )

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='set_ok')


class InstanceStopExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'stop_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='stopped',
                erred_state='erred',
            ),
        )


class InstanceStartExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'start_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='running',
                erred_state='erred',
            ),
            core_tasks.BackendMethodTask().si(serialized_instance, 'pull_instance_public_ips'),
        )


class InstanceRestartExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'reboot_instance', state_transition='begin_updating',
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_instance,
                backend_pull_method='pull_instance_runtime_state',
                success_state='running',
                erred_state='erred',
            ),
        )


class InstanceDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if not instance.backend_id:
            return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_deleting')

        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance, 'destroy_instance', state_transition='begin_deleting'),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance, backend_check_method='is_instance_terminated'),
        )


class AWSCleanupExecutor(structure_executors.BaseCleanupExecutor):
    executors = (
        (models.Instance, InstanceDeleteExecutor),
        (models.Volume, VolumeDeleteExecutor),
    )



import logging
import re

from waldur_slurm.base import BatchError, BaseBatchClient
from waldur_slurm.parser import SlurmReportLine
from waldur_slurm.structures import Account, Association
from waldur_slurm.utils import format_current_month


class SlurmError(BatchError):
    pass


logger = logging.getLogger(__name__)


class SlurmClient(BaseBatchClient):
    """
    This class implements Python client for SLURM.
    See also: https://slurm.schedmd.com/sacctmgr.html
    """

    def list_accounts(self):
        output = self._execute_command(['list', 'account'])
        return [self._parse_account(line) for line in output.splitlines() if '|' in line]

    def _parse_account(self, line):
        parts = line.split('|')
        return Account(
            name=parts[0],
            description=parts[1],
            organization=parts[2],
        )

    def get_account(self, name):
        output = self._execute_command(['show', 'account', name])
        lines = [line for line in output.splitlines() if '|' in line]
        if len(lines) == 0:
            return None
        return self._parse_account(lines[0])

    def create_account(self, name, description, organization, parent_name=None):
        parts = [
            'add', 'account', name,
            'description="%s"' % description,
            'organization="%s"' % organization,
        ]
        if parent_name:
            parts.append('parent=%s' % parent_name)
        return self._execute_command(parts)

    def delete_all_users_from_account(self, name):
        return self._execute_command(['remove', 'user', 'where', 'account=%s' % name])

    def account_has_users(self, account):
        output = self._execute_command([
            'show', 'association', 'where', 'account=%s' % account
        ])
        items = [self._parse_association(line) for line in output.splitlines() if '|' in line]
        return any(item.user != '' for item in items)

    def delete_account(self, name):
        if self.account_has_users(name):
            self.delete_all_users_from_account(name)

        return self._execute_command(['remove', 'account', 'where', 'name=%s' % name])

    def set_resource_limits(self, account, quotas):
        quota = 'GrpTRES=cpu=%d,gres/gpu=%d,mem=%d' % (quotas.cpu, quotas.gpu, quotas.ram)
        return self._execute_command(['modify', 'account', account, 'set', quota])

    def get_association(self, user, account):
        output = self._execute_command([
            'show', 'association', 'where', 'user=%s' % user, 'account=%s' % account
        ])
        lines = [line for line in output.splitlines() if '|' in line]
        if len(lines) == 0:
            return None
        return self._parse_association(lines[0])

    def _parse_association(self, line):
        parts = line.split('|')
        value = parts[9]
        match = re.match(r'cpu=(\d+)', value)
        if match:
            value = int(match.group(1))
        return Association(
            account=parts[1],
            user=parts[2],
            value=value,
        )

    def create_association(self, username, account, default_account=''):
        return self._execute_command(['add', 'user', username,
                                      'account=%s' % account,
                                      'DefaultAccount=%s' % default_account])

    def delete_association(self, username, account):
        return self._execute_command([
            'remove', 'user', 'where', 'name=%s' % username, 'and', 'account=%s' % account
        ])

    def get_usage_report(self, accounts):
        month_start, month_end = format_current_month()

        args = [
            '--noconvert',
            '--truncate',
            '--allocations',
            '--allusers',
            '--starttime=%s' % month_start,
            '--endtime=%s' % month_end,
            '--accounts=%s' % ','.join(accounts),
            '--format=Account,ReqTRES,Elapsed,User',
        ]
        output = self._execute_command(args, 'sacct', immediate=False)
        return [SlurmReportLine(line) for line in output.splitlines() if '|' in line]

    def _execute_command(self, command, command_name='sacctmgr', immediate=True):
        account_command = [command_name, '--parsable2', '--noheader']
        if immediate:
            account_command.append('--immediate')
        account_command.extend(command)
        return self.execute_command(account_command)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-27 08:31
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('waldur_vmware', '0003_disk'),
    ]

    operations = [
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('guest_os', models.CharField(help_text='Defines the valid guest operating system types used for configuring a virtual machine', max_length=50)),
                ('cores', models.PositiveSmallIntegerField(default=0, help_text='Number of cores in a VM')),
                ('cores_per_socket', models.PositiveSmallIntegerField(default=1, help_text='Number of cores per socket in a VM')),
                ('ram', models.PositiveIntegerField(default=0, help_text='Memory size in MiB', verbose_name='RAM')),
                ('created', models.DateTimeField()),
                ('modified', models.DateTimeField()),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
    ]



from django.contrib.auth import get_user_model

from waldur_core.logging.loggers import EventLogger, event_logger


User = get_user_model()

provider_event_type_mapping = {
    'google': 'auth_logged_in_with_google',
    'facebook': 'auth_logged_in_with_facebook',
    'smartid.ee': 'auth_logged_in_with_smartid_ee',
    'tara': 'auth_logged_in_with_tara',
}


class SocialEventLogger(EventLogger):
    provider = str
    user = User

    class Meta:
        event_types = provider_event_type_mapping.values()
        event_groups = {'users': event_types}

    @staticmethod
    def get_scopes(event_context):
        return {event_context['user']}


event_logger.register('auth_social', SocialEventLogger)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-18 11:00
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.logging.loggers
import waldur_core.structure.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0009_project_is_removed'),
    ]

    operations = [
        migrations.CreateModel(
            name='VMwareService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'verbose_name': 'VMware provider',
                'verbose_name_plural': 'VMware providers',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VMwareServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_vmware.VMwareService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'VMware provider project link',
                'verbose_name_plural': 'VMware provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='vmwareservice',
            name='projects',
            field=models.ManyToManyField(related_name='_vmwareservice_projects_+', through='waldur_vmware.VMwareServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='vmwareservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AlterUniqueTogether(
            name='vmwareserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='vmwareservice',
            unique_together=set([('customer', 'settings')]),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-30 10:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0002_offeringcomponent_use_limit_for_billing'),
    ]

    operations = [
        migrations.AddField(
            model_name='componentusage',
            name='recurring',
            field=models.BooleanField(default=False, help_text='Reported value is reused every month until changed.'),
        ),
    ]



from rest_framework import serializers

from waldur_core.core import serializers as core_serializers


class ApplicationSerializerRegistry:
    """
    Holds application-related model-serializer mappings
    """
    APPLICATION_SERIALIZERS = dict()

    @staticmethod
    def get_registered_app_serializer(model_class):
        return ApplicationSerializerRegistry.APPLICATION_SERIALIZERS[model_class]

    @staticmethod
    def register(model_class, application_serializer):
        ApplicationSerializerRegistry.APPLICATION_SERIALIZERS[model_class] = application_serializer


class ApplicationSerializerMetaclass(serializers.SerializerMetaclass):
    """ Registers model-serializer mapper for various types of application serializers
    """

    def __new__(cls, name, bases, args):
        serializer = super(ApplicationSerializerMetaclass, cls).__new__(cls, name, bases, args)
        ApplicationSerializerRegistry.register(args['Meta'].model, serializer)
        return serializer


class BaseApplicationSerializer(core_serializers.AugmentedSerializerMixin,
                                serializers.HyperlinkedModelSerializer,
                                metaclass=ApplicationSerializerMetaclass):
    class Meta:
        model = NotImplemented


class SummaryApplicationSerializer(core_serializers.BaseSummarySerializer):
    @classmethod
    def get_serializer(cls, model):
        return ApplicationSerializerRegistry.get_registered_app_serializer(model)



from django.apps import AppConfig
from django.db.models import signals


class EventsConfig(AppConfig):
    name = 'waldur_core.logging'
    verbose_name = 'Logging'

    def ready(self):
        from waldur_core.logging import handlers, models

        signals.post_save.connect(
            handlers.process_hook,
            sender=models.Event,
            dispatch_uid='waldur_core.logging.handlers.process_hook',
        )



from django.contrib.contenttypes.models import ContentType

from waldur_core.core.utils import DryRunCommand
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace_openstack import utils
from waldur_openstack.openstack_tenant import models as tenant_models


class Command(DryRunCommand):
    help = """Pull OpenStack volumes metadata to marketplace."""

    def handle(self, dry_run, *args, **options):
        content_type = ContentType.objects.get_for_model(tenant_models.Volume)
        resources = marketplace_models.Resource.objects.filter(content_type=content_type).exclude(object_id=None)
        for resource in resources:
            if resource.scope:
                utils.import_volume_metadata(resource)
        self.stdout.write(self.style.SUCCESS('%s resources have been processed.' % resources.count()))



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models


class Invitation(core_models.UuidMixin,
                 TimeStampedModel,
                 core_models.ErrorMessageMixin,
                 core_models.UserDetailsMixin):
    class Permissions:
        customer_path = 'customer'

    class State:
        REQUESTED = 'requested'
        REJECTED = 'rejected'
        PENDING = 'pending'
        ACCEPTED = 'accepted'
        CANCELED = 'canceled'
        EXPIRED = 'expired'

        CHOICES = (
            (REQUESTED, 'Requested'),
            (REJECTED, 'Rejected'),
            (PENDING, 'Pending'),
            (ACCEPTED, 'Accepted'),
            (CANCELED, 'Canceled'),
            (EXPIRED, 'Expired'),
        )

    created_by = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, related_name='+', blank=True, null=True)
    approved_by = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, related_name='+', blank=True, null=True)

    customer = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.Customer, verbose_name=_('organization'), related_name='invitations')
    customer_role = structure_models.CustomerRole(verbose_name=_('organization role'), null=True, blank=True)

    project = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.Project, related_name='invitations', blank=True, null=True)
    project_role = structure_models.ProjectRole(null=True, blank=True)

    state = models.CharField(max_length=10, choices=State.CHOICES, default=State.PENDING)
    link_template = models.CharField(max_length=255, help_text=_('The template must include {uuid} parameter '
                                                                 'e.g. http://example.com/invitation/{uuid}'))
    email = models.EmailField(help_text=_('Invitation link will be sent to this email. Note that user can accept '
                                          'invitation with different email.'))
    civil_number = models.CharField(
        max_length=50, blank=True,
        help_text=_('Civil number of invited user. If civil number is not defined any user can accept invitation.'))
    tax_number = models.CharField(_('tax number'), max_length=50, blank=True)

    def get_expiration_time(self):
        return self.created + settings.WALDUR_CORE['INVITATION_LIFETIME']

    def accept(self, user):
        if self.project_role is not None:
            self.project.add_user(user, self.project_role, self.created_by)
        else:
            self.customer.add_user(user, self.customer_role, self.created_by)

        self.state = self.State.ACCEPTED
        self.save(update_fields=['state'])

    def cancel(self):
        self.state = self.State.CANCELED
        self.save(update_fields=['state'])

    def approve(self, user):
        self.state = self.State.PENDING
        self.approved_by = user
        self.save(update_fields=['state', 'approved_by'])

    def reject(self):
        self.state = self.State.REJECTED
        self.save(update_fields=['state'])

    def __str__(self):
        return self.email
